import argparse
from dataclasses import dataclass
import os
import shutil
import time

from lspinstaller.config import load_config
from lspinstaller.constants import LSP_HOME
from lspinstaller.resolver.resolver import resolve
from .data import sources

class Arguments(argparse.Namespace):
    packages: list[str] | None


def run_list(_args):
    print(f"There are {len(sources)} available servers.")
    print()
    print('   {0:<30} How installed'.format("Name in lspinstaller"))
    for (name, data) in sources.items():
        builder = f" * {name:<30} "
        if ("github" in data):
            builder += f"via GitHub: {data['github']['fragment']}"
        elif ("npm" in data):
            builder += f"via npm: {data['npm']['package']}"
        else:
            builder += "Unknown source. Someone added a type without updating list"
        print(builder)

def run_install(args):
    assert args.packages is not None
    print(f"Installing: {', '.join(args.packages)}")

    config = load_config()

    for package in args.packages:
        if not package in sources:
            print(
                f"{package} is not a known lsp. See lspinstaller list for the available LSPs"
            )
            exit(-1)
        elif package in config.packages:
            print(
                f"{package} has already been installed. Use update instead"
            )
            exit(-1)

    # This is not inlined into the previous for loop intentionally. 
    # The previous loop is just validation, and gives much more rapid feedback
    # than a unified loop would be able to do.
    for package in args.packages:
        spec = sources[package]

        if "binary" in spec:
            output_path = os.path.join(
                LSP_HOME,
                package
            )
            if os.path.exists(
                output_path
            ):
                print(f"WARNING: {output_path} already exists.")
                print("This is likely a weird state artefact. The folder will now be removed before the install is attempted")
                print("If this is a mistake, press CTRL-C to abort NOW!")

                time.sleep(10)

                shutil.rmtree(
                    output_path
                )

        resolve(package, spec, config)
    config.commit()


def run_update(_args):
    pass

def start_cli():
    parser = argparse.ArgumentParser(
        prog="lspinstaller",
        description="Installs (some) LSP servers",
        epilog="lspinstaller is licensed under the MIT license: https://codeberg.org/LunarWatcher/lspinstaller/src/branch/master/LICENSE",
    )

    subs = parser.add_subparsers(required=True)
    cmd_list = subs.add_parser(
        "list",
        help="List the available LSP servers"
    )
    cmd_list.set_defaults(func=run_list)

    cmd_install = subs.add_parser(
        "install",
        help="Install an LSP server"
    )
    cmd_install.add_argument(
        "packages",
        nargs="+",
        help="The packages to install",
    )
    cmd_install.set_defaults(func=run_install)

    cmd_update = subs.add_parser(
        "update",
        help="Updates all LSP servers"
    )
    cmd_update.set_defaults(func=run_update)

    args: Arguments = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    start_cli()



from dataclasses import dataclass
import os
import stat
import subprocess
import shutil
import platform
from lspinstaller.config import Config
from lspinstaller.constants import LSP_HOME
from lspinstaller.data import sources
from lspinstaller.resolver.util.github import default_version_parser, get_release_info
from lspinstaller.resolver.util.fetch import fetch

@dataclass
class Data:
    version: str
    os: str

def init_data(version: str):
    return Data(
        version=version,
        os=platform.system().lower()
    )

def parse_special(value, data: Data):
    if not isinstance(value, str):
        value(data)
    else:
        # TODO: This will not scale well. Do not add more fields this way;
        # write a proper replacement engine instead.
        value = value.replace("${os}", data.os)
        value = value.replace("${version}", data.version)

        return value

def resolve(package_name, spec, config: Config):
    assert spec is not None
    os.makedirs(
        os.path.join(LSP_HOME, "bin"),
        exist_ok=True
    )
    if "removed" in spec:
        print(
            f"{package_name} has been removed, and will not be updated or installed"
        )
    elif "github" in spec:
        release = get_release_info(
            spec["github"]["fragment"],
            spec.get("version_parser", default_version_parser)
        )
        binary = spec["binary"]
        assert binary is not None, \
            f"misconfigured binary object for {package_name}"

        if "url" in binary:
            url = binary["url"]
            url = parse_special(
                url,
                init_data(
                    release.tag_name
                )
            )
        else:
            pattern = binary["pattern"]
            assert pattern is not None
            pattern = parse_special(
                pattern,
                init_data(
                     release.tag_name
                ),
            )
            url = None
            for asset in release.assets:
                if asset.name == pattern:
                    print(f"Resolved asset to {asset.name}")
                    url = asset.url
                    break
            else:
                print(f"Failed to resolve asset for {package_name}")
                exit(-1)

        # Should never throw, but required because url is str | None due to the
        # = None assignment in the previous if statement
        assert url is not None
        root = fetch(
            LSP_HOME,
            url,
            package_name,
            binary["archive"],
            binary.get("is_nested", False)
        )

        for (dest_name, src_path) in binary["link"].items():

            full_src = os.path.join(root, src_path)
            full_dest = os.path.join(
                LSP_HOME,
                "bin",
                dest_name
            )
            if platform.system() != "Windows":
                print("Chmoding...")
                os.chmod(
                    full_src,
                    mode=os.stat(full_src).st_mode | stat.S_IEXEC
                )
            # TODO: symlinking breaks clangd
            # print(f"Symlinking {full_src} -> {full_dest}")
            # if os.path.exists(full_dest):
                # os.remove(full_dest)
            # os.link(
                # full_src, full_dest
            # )


        config.update_package(package_name, release.tag_name)
        return release.tag_name
    elif "npm" in spec:
        subprocess.run(
            ["npm", "install", spec["npm"]["package"]],
            cwd=LSP_HOME
        )
        return None

def do_update(config: Config):
    if os.path.exists(
        os.path.join(
            LSP_HOME,
            "node_modules"
        )
    ):
        # Pretty sure this should just work:tm:, since npm install uses the ^
        # notation, which should work with npm update:
        # https://stackoverflow.com/a/19824154
        subprocess.run(
            ["npm", "update"],
            cwd=LSP_HOME
        )

    for package in config.packages:
        # Nuke the existing tree just in case
        dest = os.path.join(
            LSP_HOME,
            package
        )
        if os.path.exists(dest):
            shutil.rmtree(
                dest
            )

        new_version = resolve(
            package,
            sources[package],
            config
        )
    config.commit()



from dataclasses import dataclass
from enum import Enum

"""
# Syntax
## Root
```
"package name": { data object }
```

The package name is also used for folders in some cases, so make sure the
package name is plain text!

## Data objects

Data objects can have the following main keys:

* github
* npm
* binary
    * The binary spec describes a binary file. Its exact meaning depends on the
      main provider
    * Must be included when using any of  these providers: "github".

### `github`

The github object contains the following keys:
* `fragment`: must be in the format `user/repo`

In addition, the separate `binary` object must be specified.

### `npm`

The npm object contains the following keys:
* `package`: the package name as it's used on npm

### `binary`

The binary object contains the following keys:
* `pattern`: required when using `github` or another provider that finds
  multiple files. Otherwise optional. Mutually exclusive with `url`
* `url`: required when the object is used with a source that's only used for
  the version. See kotlin-lsp for an example. Mutually exclusive with `pattern`
* `link`: describes which files to symlink to the special bin directory.
  Syntax: `"file name in lsp/bin": "filename relative to download root"`
  Currently does nothing due to a severe bug where symlinks break the binary.
* `archive`: the archive type. Required. Accepted values: "zip".
* `is_nested`: whether or not the archive is nested, meaning the main archive
  has an inner folder. Clangd, for example, has this, where extracting the
  clangd archive extracts another clangd-<version> folder. Clangd therefore
  sets this to true. kotlin-lsp, on the other hand, just has its bin directory
  immediately after extraction, so it sets false. Required if `archive` is set

The binary object is special, and will never be executed standalone. It exists
to consume data from other providers to actually do the install.

## Special values

In some strings, there are special values. These are in the format `${key}`,
and the currently recognised values are:

* `os`: One of the literals `windows` or `linux`
* `version`: the version as determined by a provider. Note that for versions
  including the `v` prefix, the `v` is stripped and must be included literally
  in the resulting string.
"""
# TODO: need a better way to manage the logic for some of these variables,
# because they are not universally portable. Just with operating systems,
# upper-case first letter does occur in some projects. Some also abbreviate
# windows. Can probably sneak in more lambdas to mitigate this, but  that
# requires passing the whole data object  to the lambda, which is just a mess.
# Lambdas in python are also fairly shit, so that wouldn't be pretty. Maybe
# there's a light-weight templating language that makes sense to use?
sources = {
    "clangd": {
        "github": {
            "fragment": "clangd/clangd",
        },
        "binary": {
            "pattern": "clangd-${os}-${version}.zip",
            "link": {
                "clangd": "bin/clangd"
            },
            "archive": "zip",
            "is_nested": True
        }
    },
    "tsserver": {
        "npm": {
            "package": "typescript-language-server"
        }
    },
    "pyright": {
        "npm": {
            "package": "pyright"
        }
    },
    "kotlin-lsp": {
        "github": {
            "fragment": "Kotlin/kotlin-lsp"
        },
        "binary": {
            "url": "https://download-cdn.jetbrains.com/kotlin-lsp/${version}/kotlin-${version}.zip",
            "link": {
                "kotlin-lsp": "kotlin-lsp.sh"
            },
            "archive": "zip",
            "is_nested": False
        },
        "version_parser": lambda raw : raw[raw.index("v") + 1:]
    }
}
